<?php

namespace cn\com\maiyoule\mqttclient;

use cn\com\maiyoule\mqttclient\biz\client\SendMQTTMessage;
use Psr\Log\LoggerInterface;
use Workerman\Mqtt\Client;
use Workerman\Worker;

/**
 * MQTT 客户端封装,用于主管理员
 */
class AuthorAppMQTTClient
{
    private static ?self $instance = null;

    public static function getInstance(): self
    {
        if (self::$instance == null) {
            self::$instance = new self();
        }
        return self::$instance;
    }

    private string $appid = '';
    private string $server = 'mqtt://mqttauthor.maiyoule.com.cn';
    private int $port = 1883;
    private string $clientId = '';
    private string $username = '';
    private string $password = '';


    private ?Client $client = null;

    public function getServer(): string
    {
        return $this->server;
    }

    public function setServer(string $server): AuthorAppMQTTClient
    {
        $this->server = $server;
        return $this;
    }

    public function getPort(): int
    {
        return $this->port;
    }

    public function setPort(int $port): AuthorAppMQTTClient
    {
        $this->port = $port;
        return $this;
    }

    public function getClientId(): string
    {
        return $this->clientId;
    }

    public function setClientId(string $clientId): AuthorAppMQTTClient
    {
        $this->clientId = $clientId;
        return $this;
    }

    public function getUsername(): string
    {
        return $this->username;
    }

    public function setUsername(string $username): AuthorAppMQTTClient
    {
        $this->username = $username;
        return $this;
    }

    public function getPassword(): string
    {
        return $this->password;
    }

    public function setPassword(string $password): AuthorAppMQTTClient
    {
        $this->password = $password;
        return $this;
    }


    public function getAppid(): string
    {
        return $this->appid;
    }

    public function setAppid(string $appid): AuthorAppMQTTClient
    {
        $this->appid = $appid;
        return $this;
    }


    private LoggerInterface $logger;

    public function getLogger(): LoggerInterface
    {
        return $this->logger;
    }

    public function setLogger(LoggerInterface $logger): void
    {
        $this->logger = $logger;
    }


    /**
     * @throws \Exception
     */
    private function __construct()
    {
        if (class_exists('Worker')) {
            throw new \Exception('没有找到Workman');
        }
    }

    private ?Worker $worker;

    /**
     * 启动服务
     * @return bool
     * @throws \Throwable
     */
    public function start(): bool
    {
        $this->worker = new Worker();

        $this->worker->onWorkerStart = function () {
            $this->logger->info('Worker 启动');
            $this->__startWorker();
        };
        $this->worker->onMessage = function ($conn, $data) {
            $this->logger->debug($data);
        };
        $this->worker->onWorkerStop = function () {
            $this->logger->info('Worker 停止');
            $this->__stopWorker();
        };
        $this->worker->onError = function () {
            $this->logger->error('Worker启动遇到错误');
        };
        if (!Worker::isRunning()) {
            Worker::runAll();
        }
        return true;
    }

    /**
     * 停止
     * @return void
     */
    public function stop(): void
    {
        if (is_null($this->worker)) {
            return;
        }
        $this->worker->stop();
        $this->worker = null;
    }

    private function __stopWorker(): void
    {
        if (is_null($this->client)) {
            return;
        }
        $topic = sprintf('biz/%s/#', $this->getAppid());

        $this->client->unsubscribe($topic);
        //清理订阅
        $this->client->disconnect();
    }

    /**
     * @throws \Exception
     */
    private function __startWorker(): void
    {
        //参考 https://www.workerman.net/doc/workerman/components/workerman-mqtt.html
        $option = [
            'keepalive' => 29,
            'client_id' => $this->getClientId(),
            'protocol_name' => 'MQTT',
            'protocol_level' => 5,
            'clean_session' => true,
            'reconnect_period' => 2,
            'connect_timeout' => 10,
            'username' => $this->getUsername(),
            'password' => $this->getPassword(),
            'resubscribe' => true,
            'debug' => false
        ];
        $address = sprintf('mqtt://%s:%d', $this->getServer(), $this->getPort());
        $this->logger->debug('建立MQTT连接到' . $address);

        $this->client = new Client($address, $option);
        //连接回调
        $this->client->onConnect = function (Client $connection) {
            $this->logger->debug('MQTT 连接成功');
            $topic = sprintf('biz/%s/#', $this->getAppid());
            $connection->subscribe($topic, ['qos' => 0], function (\Exception|null $exception, array $granted) {

                if (!is_null($exception)) {
                    $this->logger->error($exception->getMessage());
                }
                $this->logger->debug('主题订阅:' . json_encode($granted));
            });

            $this->callEvent(self::EVENT_CONNECT, $connection);
        };
        $this->client->onClose = function () {
            $this->logger->debug('MQTT 连接断开');
            $this->callEvent(self::EVENT_CLOSE);
        };
        $this->client->onMessage = function (string $topic, string $data, Client $client) {
            //收到消息
            $this->logger->debug(sprintf('MQTT 收到消息 Topic:%s Data:%s', $topic, $data));
            $this->callEvent(self::EVENT_MESSAGE, $topic, $data, $client);
        };
        //遇到错误
        $this->client->onError = function (\Exception $err) {
            $this->logger->error('MQTT 遇到错误' . $err->getMessage());
            $this->callEvent(self::EVENT_ERROR, $err);
        };
        $this->client->connect();

    }

    const EVENT_CONNECT = 'connect';
    const EVENT_CLOSE = 'close';

    const EVENT_MESSAGE = 'message';
    const EVENT_ERROR = 'error';

    private array $mapEvent = [];

    public function registerEvent(string $event, \Closure $callback): AuthorAppMQTTClient
    {
        $this->mapEvent[$event][] = $callback;
        return $this;
    }

    private function callEvent(string $event, ...$args): void
    {
        if (isset($this->mapEvent[$event])) {
            foreach ($this->mapEvent[$event] as $callback) {
                call_user_func_array($callback, $args);
            }
        }
    }

    /**
     * 发送消息到目标客户端
     * @param SendMQTTMessage $message
     * @return bool
     */
    public function sendToTarget(SendMQTTMessage $message): bool
    {
        //组装主题
        $topic = sprintf('biz/%s/%s/%s/downstream', $this->getAppid(), $message->getBiz(), $message->getTargetId());
        if (is_null($this->client)) {
            return false;
        }
        $options = [
            'qos' => $message->getQoS(),
            'retain' => $message->isRetain(),
            'dup' => false
        ];
        $this->client->publish($topic, $message->getBiz(), $options);
        return true;
    }


}